# Copyright (C) 2007-2008 www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/

"""Cross platform library to locate the desktop folder."""

import os
import re
import sys

USER_FOLDER = os.path.expanduser('~')

if sys.platform.startswith('win'):
    # Windows
    from win32com.shell import shell, shellcon
    DESKTOP_FOLDER = shell.SHGetFolderPath(0, shellcon.CSIDL_DESKTOP,
        None, 0)
elif sys.platform.startswith('darwin'):
    # Mac: verify this!
    DESKTOP_FOLDER  = os.path.expanduser('~/Desktop')
else:
    # Linux
    user_dirs  = os.path.expanduser('~/.config/user-dirs.dirs')
    if os.path.exists(user_dirs):
        DESKTOP_FOLDER = os.path.expanduser(
            re.search('XDG_DESKTOP_DIR="(.*?)"',open(user_dirs).read())\
                .group(1).replace('$HOME','~'))
    else:
        DESKTOP_FOLDER  = os.path.expanduser('~/Desktop')
    del user_dirs
    
if not os.path.isdir(DESKTOP_FOLDER):
    DESKTOP_FOLDER  = USER_FOLDER
    
if __name__ == '__main__':
    sys.stdout.write('Your desktop is: %s\n'%DESKTOP_FOLDER)
