# Phatch - Photo Batch Processor
# Copyright (C) 2007-2008 www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/
#
# Phatch recommends SPE (http://pythonide.stani.be) for editing python files.

"""All PIL related issues."""

#License: latest GPL
import datetime, os, re, time, types
from cStringIO import StringIO
from urllib2 import urlopen

import EXIF

import Image
try:
    from ExifTags import TAGS as EXIFTAGS
except:
    #older versions of PIL
    EXIFTAGS    = {}
    
try:
    import pyexiv2
    import _pyexiv2 as exif
except:
    pyexiv2     = None
    exif        = False
WWW_PYEXIV2     = 'http://tilloy.net/dev/pyexiv2/'
NEEDS_PYEXIV2   = _('pyexiv2 needs to be installed')+' (%s)'%WWW_PYEXIV2

try:
    from unicoding import ensure_unicode
    from core.ct import TITLE
    from core.translation import _t
    from core.config import USER_BIN_PATH
    from formField import is_www_file
except ImportError:
    def ensure_unicode(x): return x
    TITLE           = ''
    _t              = str
    USER_BIN_PATH   = os.path.expanduser('~/.phatch/bin')
    is_www_file     = os.path.isfile

from core.lib import libtiff
if not libtiff.init([USER_BIN_PATH]):
    libtiff         = None
    
EXT_BY_FORMATS = {
    'GIF'   : ['gif'],
    'JPEG'  : ['jpg','jpeg','jpe'],
    'PNG'   : ['png'],
    'TIFF'  : ['tif','tiff'],
}
FORMATS_BY_EXT = {}
for format, exts in EXT_BY_FORMATS.items():
    for ext in exts:
        FORMATS_BY_EXT[ext] = format

IMAGE_DEFAULT_DPI       = 72
DOT         = '.' #should be same as in core.translations
MONTHS      = (_t('january'),_t('february'),_t('march'),_t('april'),
                _t('may'),_t('june'),_t('july'),_t('august'),
                _t('september'),_t('october'),_t('november'),
                _t('december'),)
WEEKDAYS    = (_t('monday'),_t('tuesday'),_t('wednesday'),
                _t('thursday'),_t('friday'),_t('saturday'),_t('sunday'))
DATETIME_KEYS   = ['year','month','day','hour','minute','second']
re_DATETIME = re.compile(
                '(?P<year>\d{4})[-:](?P<month>\d{2})[-:](?P<day>\d{2}) '
                '(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})')
                
re_TAG      = re.compile('(Pil|Exif|Iptc|Pexif|Zexif)([.]\w+)+')
re_KEY      = re.compile('(#*)((\w|[.])*)')
WWW_CACHE   = {}

MODULES     = []
if libtiff:
    MODULES.append(libtiff)
    
def get_format(ext):
    return FORMATS_BY_EXT.get(ext.lower(),ext.upper())
    
def check_libtiff(compression):
    if not(libtiff or compression in ['raw','None']):
        raise Exception('Libtiff tools are needed for "%s" compression'\
            %compression)
            
def transpose(image, methods):
    """Transpose with a sequence of transformations, mainly useful
    for exif."""
    for method in methods:
        image = image.transpose(method)
    return image
            
def treshold(a):
    if a <= 128:
        return 255
    else:
        return 0

def image_open(x):
    if is_www_file(x):
        try:
            return WWW_CACHE[x]
        except KeyError:
            f   = urlopen(x)
            im  = WWW_CACHE[x] = Image.open(StringIO(f.read()))
            return im
    return Image.open(x)

def extract_info_dpi(info,image):
    #info
    image_info              = image.info
    if image_info.has_key('dpi'):
        #also in this case it is possible that dpi is 0
        dpi                 = image_info['dpi']
        if len(dpi) > 1:
            dpi             = dpi[0]
        info[_t('dpi')]  = dpi
    else: dpi               = 0
    if dpi == 0:
        if image_info.has_key('aspect'):
            info[_t('dpi')]  = dpi = image_info['aspect'][0]
        else:
            dpi             = 0
        if dpi == 0:
            info[_t('dpi')]  = IMAGE_DEFAULT_DPI
        
def split_data(new):
    """new - dictionary"""
    value   = new.values()[0]
    #tuples or list
    if type(value) in (types.ListType,types.TupleType):
        if len(value) > 1:
            for k, v in new.items():
                for i, x in enumerate(v):
                    new['%s%s%d'%(k,DOT,i)]  = v[i]
        return
    #datetime strings
    done    = False
    for k, v in new.items():
        if type(v) in types.StringTypes:
            dt      = re_DATETIME.match(v)
            if dt:
                for key in DATETIME_KEYS:
                    new['%s%s%s'%(k,DOT,key)]  = dt.group(key)
                    done    = True
    if done: return 
    #date time values
    if type(value) == datetime.datetime:
        for k, v in new.items():
            for key in DATETIME_KEYS:
                new['%s%s%s'%(k,DOT,key)]  = getattr(v,key)
        
def extract_info_pexif(info,image):
    """pexif = Pil EXIF"""
    #check if exif information is available through pil
    if not hasattr(image,'_getexif'):
        return info
    #add exif information
    try:
        exif    = image._getexif()
    except:
        exif    = None
    if not exif: return
    for key, value in exif.items():
        new    = {'Zexif%s0x%04x'%(DOT,key):value}
        if EXIFTAGS.has_key(key):
            new['Pexif%s%s'%(DOT,EXIFTAGS[key])] = value
        #Check if we can split exif data further
        split_data(new)
        info.update(new)

def extract_info_pyexiv2(info,location=''):
    if not (pyexiv2 and os.path.isfile(location)):
        return
    #official patch of pyexiv2
    try:
        if type(location) is unicode:
            location = location.encode('utf-8')
    except UnicodeEncodeError:
        return
    try:
        image   = pyexiv2.Image(location)
        image.readMetadata()
    except:
        return
    #exif & iptc
    for keys in [image.exifKeys(),image.iptcKeys()]:
        for key in keys:
            try:
                new = {key:image[key]}
                split_data(new)
                info.update(new)
            except:
                pass
                
def fix_EXIF(tag):
    if not tag.startswith('EXIF'):
        tag = 'EXIF.'+tag
    return tag.replace(' ',DOT)
    
def extract_info_EXIF(info,location=''):
    if not os.path.isfile(location):
        return
    f       = open(location, 'rb')
    try:
        tags    = EXIF.process_file(f)
    except:
        tags    = {}
    f.close()
    for tag, value in tags.items():
        if tag != 'EXIF UserComment':
            info[fix_EXIF(tag)]  = value

def extract_info_location(info,location='',folder=None):
    if not location.strip():
        return
    location                    = ensure_unicode(location)
    dirname                     = os.path.dirname(location)
    if folder:
        info[_t('folder')]      = folder
    else:
        info[_t('folder')]      = dirname
    info[_t('root')],\
    info[_t('foldername')]      = os.path.split(info[_t('folder')])
    info[_t('subfolder')]       = dirname[len(info[_t('folder')])+1:]
    info[_t('path')]            = location
    filename                    = os.path.basename(location)
    info[_t('filename')], typ   = os.path.splitext(filename)
    info[_t('type')]            = typ[1:]
    #os time
    try:
        file_stat               = os.stat(location)
        info[_t('filesize')]    = file_stat.st_size
        st_mtime                = time.localtime(file_stat.st_mtime)[:7]
    except:
        info[_t('filesize')]    = 0
        st_mtime                = (0,)*7
    info[_t('year')],info[_t('month')],info[_t('day')],\
        info[_t('hour')],info[_t('minute')],info[_t('second')],\
        info[_t('weekday')]     = st_mtime
    info[_t('weekdayname')]     = WEEKDAYS[info[_t('weekday')]]
    info[_t('monthname')]       = MONTHS[info[_t('month')]-1]

def extract_info_image(info,image):
    extract_info_dpi(info,image)
    info['compression'] = image.info.get('compression','none')
    info['transparency'] = image.info.get('transparency',None)
    #load Pil info
    for key, value in image.info.items():
        if key!='exif':
            info['Pil'+DOT+key.title()] = value
    #summarize all photo properties in info
    info.update({
        'Pil'+DOT+'Format'              : image.format,
        'Pil'+DOT+'FormatDescription'   : image.format_description,
    })
    return info
    
def extract_info_geometry(info,image):
    #retrieve properties
    width, height           = size = image.size
    info.update({
        _t('width')                     : width,
        _t('height')                    : height,
        _t('mode')                      : image.mode,
        'Pil'+DOT+'Size'                : size,
    })
    return info

def extract_info_pil(image):
    return prefix_dict(extract_info_image({},image),'new'+DOT)
    
def extract_info(info, image, location='',image_index=0,folder=None,
        metadata=True,format=None):
    info['#']   = image_index
    extract_info_location(info,location,folder)
    extract_info_image(info,image)
    if format:
        info['Pil.Format'] = format
    if metadata:
        extract_info_metadata(info,location,image)
    return info
    
def extract_info_metadata(info,location,image=None):
    if image:
        extract_info_pexif(info,image)
    extract_info_pyexiv2(info,location)
    extract_info_EXIF(info,location)
    
def get_exif_transposition(image):
    orientation = get_exif_orientation(image)
    #see EXIF.py 
    if orientation == 1:
        transposition           = transposition_reverse   = ()
    elif orientation == 2:
        transposition           = Image.FLIP_LEFT_RIGHT,
        transposition_reverse   = Image.FLIP_LEFT_RIGHT,
    elif orientation == 3:
        transposition           = Image.ROTATE_180,
        transposition_reverse   = Image.ROTATE_180,
    elif orientation == 4:
        transposition           = Image.FLIP_TOP_BOTTOM,
        transposition_reverse   = Image.FLIP_TOP_BOTTOM,
    elif orientation == 5:
        transposition           = Image.FLIP_LEFT_RIGHT,\
                                        Image.ROTATE_90
        transposition_reverse   = Image.ROTATE_270,\
                                        Image.FLIP_LEFT_RIGHT
    elif orientation == 6:
        transposition           = Image.ROTATE_270,
        transposition_reverse   = Image.ROTATE_90,
    elif orientation == 7:
        transposition           = Image.FLIP_LEFT_RIGHT,\
                                        Image.ROTATE_270
        transposition_reverse   = Image.ROTATE_90,\
                                        Image.FLIP_LEFT_RIGHT
    elif orientation == 8:
        transposition           = Image.ROTATE_90,
        transposition_reverse   = Image.ROTATE_270,
    else:
        transposition           = transposition_reverse   = ()
    return transposition, transposition_reverse
    
def exif_transpose(image):
    return transpose(image,get_exif_transposition(image)[0])
            
def get_exif_orientation(image):
    if not hasattr(image,'_getexif'):
        return 1
    try:
        _exif = image._getexif()
        if not _exif:
            return 1
        return _exif[0x0112]
    except KeyError:
        return 1

def prefix_dict(d,prefix,exclude=[]):
    exclude.append(prefix)
    for key, value in d.items():
        add = True
        for e in exclude:
            if key[:len(e)] == e:
                add = False
                break
        if add:
            d[prefix+key] = value
    return d

class TestInfo:
    def __init__(self,info):
        self._keys   = info.keys()
        
    def __getitem__(self,x):
        key_match   = re_KEY.match(x)
        if key_match:
            key     = key_match.group(2)
            if not key: key = '#'
        else:
            key     = x
        if self._is_valid_key(key):
            return '2'
        raise KeyError, x
    
    def _is_valid_key(self,x):
        return x in self._keys or re_TAG.match(x)
    
class Info:
    def __init__(self,dictionary=None):
        if dictionary is None:
            self._dictionary= {'#':0}
        else:
            self._dictionary= dictionary
        self.__setitem__    = self._dictionary.__setitem__
        self.keys           = self._dictionary.keys
        self.items          = self._dictionary.items
        self.update         = self._dictionary.update
        self.values         = self._dictionary.values
        
    def __getitem__(self,x):
        match               = re_KEY.match(x)
        if match:
            number          = match.group(1)
            if number:
                key         = match.group(2)
                if not key: key = '#'
                format      = '%%0%dd'%len(number)
                return format%self._dictionary[key]
        return self._dictionary[x]
    
    def copy(self):
        return Info(self._dictionary.copy())
    
def create_test_info():
    global TEST_INFO
    image       = Image.new('L',(1,1))
    info        = extract_info({},image,'/test/unknown.png')
    info        = extract_info_geometry(info,image)
    TEST_INFO   = TestInfo(info)
    return TEST_INFO
       
def report_invalid_image(filename,valid,invalid,folder):
    try:
        im  = Layer(filename, load=False)
        im.verify()
        valid.append((folder,filename))
    except:
        invalid.append(filename)
            
class Photo:
    def __init__(self,filename,image_index=0,copy_metadata = True,
            folder=None):
        #validate image
        name = self.current_layer_name  = _t('background')
        layer                           = Layer(filename,load=True)
        self.layers                     = {name:layer}
        self.log                        = ''
        self._exif_transposition_orientation = None
        self._exif_transposition_reverse= None
        self.metadata                   = {}
        self.init_info(layer,filename,image_index,copy_metadata,folder)
        
    def flush_metadata(self):
        #is there something to be flushed?
        if not self.metadata:
            return
        #throw an error if pyexiv2 is not installed
        if not exif:
            raise ImportError(NEEDS_PYEXIV2)
        location = self.info['path']
        self.log += exif.flush(location,self.metadata,
            self.info['Pil.Format'])
        #as metadata has changed, use new source
        self.metadata       = {}
        #update
        try:
            image = Image.open(location)
        except:
            image = None
        extract_info_metadata(self.info,location,image)
        
    def get_filename(self,folder,filename,typ):
        self.info[_t('new')+DOT+_t('width')], \
            self.info[_t('new')+DOT+_t('height')] = \
            self.image.size
        return os.path.join(folder,'%s.%s'%(filename,typ))\
            .replace('<','%(').replace('>', ')s')%self.__dict__
            
    def init_info(self,layer,location,image_index,copy_metadata=True,
            folder=None,prefix=_t('new')+DOT):
        self.info_original  = Info()
        extract_info(self.info_original,layer.image,location,
            image_index,folder,metadata=True)
        self.rotate_exif()
        extract_info_geometry(self.info_original,layer.image)
        self.info           = prefix_dict(self.info_original,prefix)
        self.update_exif_source(location,self.info['Pil.Format'],
            copy_metadata)
        
    def update_exif_source(self,path,format,copy_metadata):
        if format:
            format = format.upper()
        if copy_metadata and exif:
            self._exif_source   = {'path':path,'format':format}
        else:
            self._exif_source   = None
        
    def get_info(self):
        """Use method for read-only"""
        return self.info.copy()
            
    def get_info_original(self):
        """Use method for read-only"""
        return self.info_original.copy()
            
    #---layers
    def get_layer(self,name=None):
        if name is None:
            name   = self.current_layer_name
        return self.layers[name]
    
    def set_layer(self,layer,name=None):
        if name is None:
            name   = self.current_layer_name
        self.layers[name]   = layer
        
    def as_image(self):
        return self.get_layer().image
        
    #---image operations affecting all layers
    def save(self,filename,format=None,copy_metadata=True,**options):
        """Saves a flattened image"""
        #todo: flatten layers
        image = self.get_layer().image
        
        #reverse exif previously applied exif orientation
        if exif.is_writable(format):
            image   = transpose(image,
                self._exif_transposition_reverse)

        #convert to rgb
        if image.mode == 'CMYK' and not (format in ['JPEG','TIFF']):
            self.log = \
                _('CMYK has been converted to RGB to save as %s.')\
                    %format
            image   = image.convert('RGB')
        try:
            if options.has_key('compression.tif'):
                compression = options['compression.tif']
                del options['compression.tif']
                check_libtiff(compression)
                self.log += libtiff.save(image,filename,
                    compression=compression, **options)
            else:
                image.save(filename,**options)
                #copy metadata if needed (problematic for tiff)
                if copy_metadata:
                    self.log += \
                        self.copy_metadata(filename,format,TITLE)
        except IOError, message:
            # clean up corrupted drawing
            if os.path.exists(filename):
                os.remove(filename)
            raise IOError(message)
        #update info
        if hasattr(options,'dpi'):
            self.set_attr(_t('dpi'),options['dpi'][0])
        extract_info(self.info,image,filename,
            folder=self.info['folder'], metadata=False, format=format)
        self.update_exif_source(filename,format,copy_metadata)
        
    def convert(self,mode,*args,**keyw):
        """Converts all layers to a different mode."""
        for layer in self.layers.values():
            if mode == 'P' and layer.image.mode == 'RGBA':
                alpha = layer.image.split()[3]
                layer.image = layer.image.convert('RGB').convert(
                    mode, colors=255, *args, **keyw
                )
                layer.image.paste(
                    255, Image.eval(alpha, treshold)
                )
                self.info['transparency'] = 255
            else:
                layer.image = layer.image.convert(mode,*args,**keyw)
        self.set_attr(_t('mode'),mode)
            
    def resize(self,size,method):
        """Resizes all layers to a different size"""
        size    = (max(1,size[0]),max(1,size[1]))
        for layer in self.layers.values():
            layer.image = layer.image.resize(size,method)
        self.set_size(size)
            
    def set_size(self,size):
        self.set_attr(_t('width'),size[0])
        self.set_attr(_t('height'),size[1])
        self.set_attr('Pil'+DOT+_t('Size'),size)
        
    def set_attr(self,attr,value):
        self.info[_t('new')+DOT+attr] = value
        
    def update_size(self):
        self.set_size(self.layers.values()[0].image.size)
        return self
        
    def rotate_exif(self,reverse=False):
        layers  = self.layers.values()
        if reverse:
            transposition   = self._exif_transposition_reverse
            self._exif_transposition_reverse = () 
        else:
            transposition, self._exif_transposition_reverse = \
                get_exif_transposition(layers[0].image)
        if transposition:
            for layer in layers:
                layer.image = transpose(layer.image,transposition)
            
    #---pil
    def apply_pil(self,function,*arg,**keyw):
        self.get_layer().apply_pil(function,*arg,**keyw)
        return self
    
    def apply_all_pil(self,function,*arg,**keyw):
        for layer in self.layers.values():
            layer.apply_pil(function,*arg,**keyw)
        return self
    
    #---exif
    def copy_metadata(self,filename,format=None,software=None):
        if self._exif_source:
            try:
                return exif.copy_metadata(
                    self._exif_source['path'], 
                    filename,
                    source_format=self._exif_source['format'],
                    target_format=format,
                    reset_orientation=self._exif_transposition_orientation,
                    software=software)
            except:
                pass
        self._exif_transposition_orientation = 0 
        return ''

class Layer:
    def __init__(self,filename,position=(0,0),load=True):
        self.open(filename)
        self.position               = position
        # VERY IMPORTANT
        # do not remove load option, otherwise libtiff.py won't work
        # correctly with group4 tiff compression
        if load:
            self.image.load()
        
    def open(self, filename):
        try:
            self.image      = image_open(filename)
            #types which PIL can open, but not load
            if self.image.info.get('compression','none')\
                    .startswith('group'):
                # tiff image with group4 compression
                check_libtiff('g4')
                self.image  = libtiff.open(filename)
        except IOError, message:
            if not is_www_file(filename):
                self.image  = self._open_with_modules(filename)
            else:
                self.image  = None
            if self.image is None:
                raise IOError(message)
                
    def _open_with_modules(self, filename):
        for module in MODULES:
            try:
                return module.open(filename)
            except:
                pass
                
    def verify(self):
        self.image.verify()
        
    def apply_pil(self,function,*arg,**keyw):
        self.image  = function(self.image,*arg,**keyw)
        

        
if __name__ == '__main__':
    create_test_info()
