#!/usr/bin/env python

# Phatch - Photo Batch Processor
# Copyright (C) 2007-2008  www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/

"""Local launch script for all platforms"""

import sys
from os.path import abspath, dirname, join
from core import config

def create_paths(relative=''):
    root    = dirname(abspath(__file__))
    def expand(x):
        return abspath(join(root,relative,x))
    paths   = {
        'PHATCH_ACTIONLISTS_PATH'   : 'actionlists',
        'PHATCH_DOCS_PATH'          : 'docs',
        'PHATCH_FONTS_PATH'         : 'fonts',
        'PHATCH_FONTS_CACHE_PATH'   : 'cache/fonts',
        'PHATCH_IMAGE_PATH'         : 'images',
        'PHATCH_LOCALE_PATH'        : 'locale',
    }
    for key, path in paths.items():
        paths[key]  = expand(path)
    paths['PHATCH_PYTHON_PATH']     = root
    return paths
    
if hasattr(sys,"frozen"):
    __file__            = sys.argv[0]
    relative            = ''
else:
    relative            = '..'
PATHS                   = create_paths(relative)

def main():
    #override paths with local paths
    config_paths    = config.init_config_paths(config_paths = PATHS)
    #start application
    import app
    app.main(config_paths, app_file = __file__)

if __name__ == '__main__':
    main()
    
